"""
.. autoclass:: revscoring.scorer_models.scorer_model.ScorerModel
    :members:

.. autoclass:: revscoring.scorer_models.scorer_model.MLScorerModel
    :members:

.. autoclass:: revscoring.scorer_models.scorer_model.ScikitLearnClassifier
    :members:
"""
import io
import pickle
import time
from datetime import datetime
from statistics import mean, stdev

import yamlconf
from sklearn.metrics import auc, roc_curve
from tabulate import tabulate

from .util import normalize_json


class ScorerModel:
    """
    A model used to score a revision based on a set of features.
    """

    def __init__(self, features, version=None, stats=None):
        """
        :Parameters:
            features : `list`(`Feature`)
                A list of `Feature` s that will be used to train the model and
                score new observations.
            version : `str`
                A string describing the version of the model.
        """
        self.features = tuple(features)
        self.version = version

    def __getattr__(self, attr):
        if attr is "version":
            return None
        else:
            raise AttributeError(attr)

    def score(self, feature_values):
        """
        Make a prediction or otherwise use the model to generate a score.

        :Parameters:
            feature_values : collection(`mixed`)
                an ordered collection of values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A `dict` of statistics
        """
        raise NotImplementedError()

    def info(self):
        """
        Returns a `dict` containing information about the model.
        """
        raise NotImplementedError()

    def format_info(self):
        """
        Returns a `str` containing information about the model.
        """
        raise NotImplementedError()

    def _validate_features(self, feature_values):
        """
        Checks the features against provided values to confirm types,
        ordinality, etc.
        """
        return [feature.validate(feature_values)
                for feature, value in zip(self.feature, feature_values)]

    def _generate_stats(self, values):
        columns = zip(*values)

        stats = tuple((mean(c), stdev(c)) for c in columns)

        return stats

    def _scale_and_center(self, values, stats):

        for feature_values in values:
            yield (tuple((val-mean)/max(sd, 0.01)
                   for (mean, sd), val in zip(stats, feature_values)))

    @classmethod
    def load(cls, f):
        """
        Reads serialized model information from a file.  Make sure to open
        the file as a binary stream.
        """
        return pickle.load(f)

    def dump(self, f):
        """
        Writes serialized model information to a file.  Make sure to open the
        file as a binary stream.
        """
        pickle.dump(self, f)

    @classmethod
    def from_config(cls, config, name, section_key='scorer_models'):
        section = config[section_key][name]

        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            class_path = section['class']
            Class = yamlconf.import_module(class_path)
            assert cls != Class

            return Class.from_config(config, name, section_key=section_key)


class MLScorerModel(ScorerModel):
    """
    A machine learned model used to score a revision based on a set of
    features.

    Machine learned models are trained and tested against labeled data.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trained = None
        self.stats = None

    def __getattr__(self, attr):
        if attr is "trained":
            return None
        else:
            raise AttributeError(attr)

    def train(self, values_labels):
        """
        Trains the model on labeled data.

        :Parameters:
            values_scores : `iterable` (( `<feature_values>`, `<label>` ))
                an iterable of labeled data Where <values_labels> is an ordered
                collection of predictive values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A dictionary of model statistics.
        """
        raise NotImplementedError()

    def test(self, values_labels):
        """
        Tests the model against a labeled data.  Note that test data should be
        withheld from from train data.

        :Parameters:
            values_labels : `iterable` (( `<feature_values>`, `<label>` ))
                an iterable of labeled data Where <values_labels> is an ordered
                collection of predictive values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A dictionary of test results.
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="scorer_models"):
        """
        Constructs a model from configuration.
        """
        section = config[section_key][name]
        if 'model_file' in section:
            return cls.load(open(section['model_file'], 'rb'))
        else:
            return cls(**{k: v for k, v in section.items() if k != "class"})


class ScikitLearnClassifier(MLScorerModel):

    def __init__(self, features, classifier_model, version=None):
        super().__init__(features, version=version)
        self.classifier_model = classifier_model
        self.stats = None

    def __getattr__(self, attr):
        if attr is "stats":
            return None
        else:
            raise AttributeError(attr)

    def train(self, values_labels):
        """

        :Returns:
            A dictionary with the fields:

            * seconds_elapsed -- Time in seconds spent fitting the model
        """
        start = time.time()

        values, labels = zip(*values_labels)

        # Fit SVC model
        self.classifier_model.fit(values, labels)
        self.trained = time.time()

        return {
            'seconds_elapsed': time.time() - start
        }

    def score(self, feature_values):
        """
        Generates a score for a single revision based on a set of extracted
        feature_values.

        :Parameters:
            feature_values : collection(`mixed`)
                an ordered collection of values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A dict with the fields:

            * predicion -- The most likely class
            * probability -- A mapping of probabilities for input classes
                             corresponding to the classes the classifier was
                             trained on.  Generating this probability is
                             slower than a simple prediction.
        """
        prediction = self.classifier_model.predict([feature_values])[0]
        labels = self.classifier_model.classes_
        probas = self.classifier_model.predict_proba([feature_values])[0]
        probability = {label: proba for label, proba in zip(labels, probas)}

        doc = {
            'prediction': prediction,
            'probability': probability
        }
        return normalize_json(doc)

    def test(self, values_labels):
        """
        :Returns:
            A dictionary of test statistics with the fields:

            * accuracy -- The mean accuracy of classification
            * table -- A truth table for classification
            * roc
                * auc -- The area under the ROC curve
        """
        values, labels = zip(*values_labels)

        scores = [self.score(feature_values) for feature_values in values]

        self.stats = {
            'table': self._label_table(scores, labels),
            'accuracy': self.classifier_model.score(values, labels),
            'roc': self._roc_stats(scores, labels,
                                   self.classifier_model.classes_)
        }
        return self.stats

    def info(self):
        return normalize_json({
            'type': self.__class__.__name__,
            'version': self.version,
            'trained': self.trained,
            'stats': self.stats
        })

    def format_info(self):
        info = self.info()
        formatted = io.StringIO()
        formatted.write("ScikitLearnClassifier\n")
        formatted.write(" - type: {0}\n".format(info.get('type')))
        formatted.write(" - version: {0}\n".format(info.get('version')))
        if isinstance(info['trained'], float):
            date_string = datetime.fromtimestamp(info['trained']).isoformat()
            formatted.write(" - trained: {0}\n".format(date_string))
        else:
            formatted.write(" - trained: {0}\n".format(info.get('trained')))

        formatted.write("\n")
        formatted.write(self.format_stats())
        return formatted.getvalue()

    def format_stats(self):
        if self.stats is None:
            return "No stats available"
        else:
            formatted = io.StringIO()
            predicted_actuals = self.stats['table'].keys()
            possible = list(set(actual for _, actual in predicted_actuals))
            possible.sort()

            formatted.write("Accuracy: {0}\n\n".format(self.stats['accuracy']))
            if 'auc' in self.stats['roc']:
                formatted.write("ROC-AUC: {0}\n\n"
                                .format(self.stats['roc']['auc']))
            else:
                formatted.write("ROC-AUC:\n")

                table_data = [[comparison_label,
                               self.stats['roc'][comparison_label]['auc']]
                              for comparison_label in possible]
                formatted.write(tabulate(table_data))
                formatted.write("\n\n")

            table_data = []

            for actual in possible:
                table_data.append(
                    [actual] +
                    [self.stats['table'].get((predicted, actual), 0)
                     for predicted in possible]
                )
            formatted.write(tabulate(table_data, headers=possible))

            return formatted.getvalue()

    @classmethod
    def _roc_stats(cls, scores, labels, possible_labels):

        if len(possible_labels) <= 2:
            # Binary classification, class choice doesn't matter.
            comparison_label = possible_labels[0]
            return cls._roc_single_class(scores, labels, comparison_label)
        else:
            roc_stats = {}
            for comparison_label in possible_labels:
                roc_stats[comparison_label] = \
                    cls._roc_single_class(scores, labels, comparison_label)

            return roc_stats

    @classmethod
    def _roc_single_class(cls, scores, labels, comparison_label):
        probabilities = [s['probability'][comparison_label]
                         for s in scores]

        true_positives = [l == comparison_label for l in labels]
        fpr, tpr, thresholds = roc_curve(true_positives, probabilities)

        return {
            'auc': auc(fpr, tpr)
        }

    @staticmethod
    def _label_table(scores, labels):

        predicteds = [s['prediction'] for s in scores]

        table = {}
        for pair in zip(labels, predicteds):
            table[pair] = table.get(pair, 0) + 1

        return table



import logging

import celery
import revscoring.errors
from celery.signals import before_task_publish

import mwapi.errors

from ..score_caches import ScoreCache
from ..metrics_collectors import MetricsCollector
from ..util import jsonify_error
from .timeout import Timeout, TimeoutError

logger = logging.getLogger(__name__)

APPLICATIONS = []


@before_task_publish.connect
def update_sent_state(sender=None, body=None, **kwargs):

    for application in APPLICATIONS:
        task = application.tasks.get(sender)
        backend = task.backend if task else application.backend

        logger.debug("Setting state to 'SENT' for {0}".format(body['id']))
        backend.store_result(body['id'], result=None, status="SENT")


class Celery(Timeout):

    def __init__(self, *args, application, **kwargs):
        super().__init__(*args, **kwargs)
        self.application = application

        expected_errors = (revscoring.errors.RevisionNotFound,
                           revscoring.errors.DependencyError,
                           mwapi.errors.RequestError,
                           mwapi.errors.TimeoutError,
                           TimeoutError)

        @self.application.task(throws=expected_errors)
        def _process_task(context, model, cache):
            return Timeout._process(self, context, model, cache)

        @self.application.task(throws=expected_errors)
        def _score_revision_task(context, model, rev_id, cache=None):
            return Timeout._score_revision(self, context, model, rev_id, cache=cache)

        APPLICATIONS.append(application)

        self._process_task = _process_task

        self._score_revision_task = _score_revision_task

    def _score_in_celery(self, context, model, rev_ids, caches):
        scores = {}
        results = {}

        if len(rev_ids) == 0:
            return scores
        if len(rev_ids) == 1:  # Special case -- do everything in celery
            rev_id = rev_ids.pop()
            id_string = self._generate_id(context, model, rev_id)
            cache = (caches or {}).get(rev_id, {})
            result = self._score_revision_task.apply_async(
                args=(context, model, rev_id), kwargs={'cache': cache},
                task_id=id_string
            )
            results[rev_id] = result
        else:  # Otherwise, try and batch
            # Get the root datasources for the rest of the batch (IO)
            root_ds_caches = self._get_root_ds(context, model, rev_ids,
                                               caches=caches)

            # Extract features and generate scores (CPU)
            for rev_id, (error, cache) in root_ds_caches.items():
                if error is not None:
                    scores[rev_id] = {'error': jsonify_error(error)}
                else:
                    id_string = self._generate_id(context, model, rev_id)
                    result = self._process_task.apply_async(
                        args=(context, model, cache),
                        task_id=id_string
                    )
                    results[rev_id] = result

        # Process async results
        for rev_id, result in results.items():
            try:
                score = result.get(self.timeout)
                scores[rev_id] = score
                self._store(context, model, rev_id, score)
            except Exception as error:
                scores[rev_id] = {'error': jsonify_error(error)}

        return scores

    def _generate_id(self, context, model, rev_id):
        scorer_model = self[context][model]
        version = scorer_model.version

        return ":".join(str(v) for v in [context, model, rev_id, version])

    def _score(self, context, model, rev_ids, caches=None):
        rev_ids = set(rev_ids)

        # Lookup scoring results that are currently in progress
        results = self._lookup_inprogress_results(context, model, rev_ids)
        missing_ids = rev_ids - results.keys()

        # Lookup scoring results that are in the cache
        scores = self._lookup_cached_scores(context, model, missing_ids)
        missing_ids = missing_ids - scores.keys()

        # Generate scores for missing rev_ids
        scores.update(self._score_in_celery(context, model, missing_ids,
                                            caches=caches))

        # Gather results
        for rev_id in results:
            try:
                scores[rev_id] = results[rev_id].get(self.timeout)
            except Exception as error:
                scores[rev_id] = {'error': jsonify_error(error)}

        # Return scores
        return scores

    def _lookup_inprogress_results(self, context, model, rev_ids):

        results = {}
        for rev_id in rev_ids:
            id_string = self._generate_id(context, model, rev_id)
            try:
                results[rev_id] = self._get_result(id_string)
            except KeyError:
                pass

        return results

    def _get_result(self, id_string):

        # Try to get an async_result for an in_progress task
        result = self._score_revision_task.AsyncResult(task_id=id_string)
        logger.debug("Checking if {0} is already being processed [{1}]"
                     .format(repr(id_string), result.state))
        if result.state not in ("SENT", "STARTED", "SUCCESS"):
            raise KeyError(id_string)
        else:
            logger.debug("Found AsyncResult for {0}".format(repr(id_string)))
            return result

    @classmethod
    def from_config(cls, config, name, section_key="score_processors"):
        logger.info("Loading Celery '{0}' from config.".format(name))
        from ..scoring_contexts import ScoringContext

        section = config[section_key][name]

        scoring_contexts = {name: ScoringContext.from_config(config, name)
                            for name in section['scoring_contexts']}

        if 'score_cache' in section:
            score_cache = ScoreCache.from_config(config, section['score_cache'])
        else:
            score_cache = None

        if 'metrics_collector' in section:
            metrics_collector = \
                MetricsCollector.from_config(config,
                                             section['metrics_collector'])
        else:
            metrics_collector = None

        timeout = section.get('timeout')
        application = celery.Celery('ores.score_processors.celery')
        application.conf.update(**{k: v for k, v in section.items()
                                   if k not in ('class', 'timeout')})

        return cls(scoring_contexts, application=application, timeout=timeout,
                   score_cache=score_cache,
                   metrics_collector=metrics_collector)



"""
``revscoring train_test -h``
::

    Trains and tests a scorer model.  This utility expects to get a file of
    tab-separated feature values and labels from which to construct a model.

    Usage:
        train_test -h | --help
        train_test <scorer_model> <features> [-p=<kv>]...
                   [--version=<vers>]
                   [--values-labels=<path>]
                   [--model-file=<path>]
                   [--label-type=<type>]

    Options:
        -h --help               Prints this documentation
        <scorer_model>          Classpath to an the MLScorerModel to construct
                                and train
        <features>              Classpath to an list of features to use when
                                constructing the model
        -p --parameter=<kv>     A key-value argument pair to use when
                                constructing the scorer_model.
        --version=<vers>        A version to associate with the model
        --values-labels=<path>  Path to a file containing feature values and
                                labels [default: <stdin>]
        --model-file=<math>     Path to write a model file to
                                [default: <stdout>]
        --label-type=<type>     Interprets the labels as the appropriate type
                                (int, float, str, bool) [default: str]
"""
import json
import random
import sys

import docopt
from tabulate import tabulate

from .util import import_from_path


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    ScorerModel = import_from_path(args['<scorer_model>'])
    features = import_from_path(args['<features>'])

    version = args['--version']

    model_kwargs = {}
    for parameter in args['--parameter']:
        key, value = parameter.split("=")
        model_kwargs[key] = json.loads(value)

    scorer_model = ScorerModel(features, version=version, **model_kwargs)

    if args['--values-labels'] == "<stdin>":
        values_labels_file = sys.stdin
    else:
        values_labels_file = open(args['--values-labels'], 'r')

    if args['--model-file'] == "<stdout>":
        model_file = sys.stdout.buffer
    else:
        model_file = open(args['--model-file'], 'wb')

    decode_label = DECODERS[args['--label-type']]

    feature_labels = read_value_labels(values_labels_file,
                                       scorer_model.features,
                                       decode_label)

    run(feature_labels, model_file, scorer_model)

DECODERS = {
    'int': lambda v: int(v),
    'float': lambda v: float(v),
    'str': lambda v: str(v),
    'bool': lambda v: v in ("True", "true", "1", "T", "y", "Y")
}


def read_value_labels(f, features, decode_label):
    for line in f:
        parts = line.strip().split("\t")
        values = parts[:-1]
        label = parts[-1]

        label = decode_label(label)

        feature_values = []
        for feature, value in zip(features, values):

            if feature.returns == bool:
                feature_values.append(value == "True")
            else:
                feature_values.append(feature.returns(value))

        yield feature_values, label


def run(feature_labels, model_file, scorer_model):

    feature_labels = list(feature_labels)
    random.shuffle(feature_labels)

    test_set_size = int(0.6*len(feature_labels))
    test_set = feature_labels[:test_set_size]
    train_set = feature_labels[test_set_size:]

    scorer_model.train(train_set)

    scorer_model.test(test_set)

    sys.stderr.write(scorer_model.format_info())

    sys.stderr.write("\n\n")

    scorer_model.dump(model_file)



import pickle

from nose.tools import eq_

from .. import hebrew
from ...datasources import revision
from ...dependencies import solve

BAD = [
    "שרמוטה"
]

INFORMAL = [
    "בגללך"  # Because of you
]

OTHER = [
    "בגלל", "חתול"
]


def compare_extraction(extractor, examples, counter_examples):

    for example in examples:
        eq_(extractor.process(example), [example])
        eq_(extractor.process("Sentence " + example + " sandwich."), [example])
        eq_(extractor.process("Sentence end " + example + "."), [example])
        eq_(extractor.process(example + " start of sentence."), [example])

    for example in counter_examples:
        eq_(extractor.process(example), [])
        eq_(extractor.process("Sentence " + example + " sandwich."), [])
        eq_(extractor.process("Sentence end " + example + "."), [])
        eq_(extractor.process(example + " start of sentence."), [])


def test_badwords():
    compare_extraction(hebrew.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(hebrew.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "סוויפט גדלה בוויומיסינג, פנסילבניה, לנאשוויל"}
    eq_(solve(hebrew.revision.words_list, cache=cache),
        ["סוויפט", "גדלה", "בוויומיסינג", "פנסילבניה", "לנאשוויל"])

    # Misspellings
    cache = {revision.text: 'בגלל חטול <td>'}
    eq_(solve(hebrew.revision.misspellings_list, cache=cache), ["חטול"])


def test_presence():
    assert hasattr(hebrew.revision, "words")
    assert hasattr(hebrew.revision, "content_words")
    assert hasattr(hebrew.revision, "badwords")
    assert hasattr(hebrew.revision, "informals")
    assert hasattr(hebrew.revision, "misspellings")

    assert hasattr(hebrew.parent_revision, "words")
    assert hasattr(hebrew.parent_revision, "content_words")
    assert hasattr(hebrew.parent_revision, "badwords")
    assert hasattr(hebrew.parent_revision, "informals")
    assert hasattr(hebrew.parent_revision, "misspellings")

    assert hasattr(hebrew.diff, "words_added")
    assert hasattr(hebrew.diff, "badwords_added")
    assert hasattr(hebrew.diff, "informals_added")
    assert hasattr(hebrew.diff, "misspellings_added")
    assert hasattr(hebrew.diff, "words_removed")
    assert hasattr(hebrew.diff, "badwords_removed")
    assert hasattr(hebrew.diff, "informals_removed")
    assert hasattr(hebrew.diff, "misspellings_removed")


def test_pickling():

    eq_(hebrew, pickle.loads(pickle.dumps(hebrew)))
